/* ---------------------------------------------------------------------- //
    bolfro.h
// ---------------------------------------------------------------------- */

#define  WINAPI  __stdcall
#define  WINAPIV __cdecl

#define  LOCALF  static __stdcall
#define  LOCALFV static __cdecl

#ifdef __cplusplus
extern "C" {
#endif

// ---------------------------------------------------------------------- //

typedef const POINT * PCPOINT;
typedef const POINTS * PCPOINTS;
typedef const RECT * PCRECT;
typedef const LOGFONT * PCLOGFONT;

#define  lft  left
#define  rgt  right
#define  btm  bottom

#define  MoveTo(hdc, x, y)  MoveToEx(hdc, x, y, NULL)

#define  RPD  (M_PI / 180.0)
#define  DPR  (180.0 / M_PI)

#define  RGBC  DWORD

typedef struct {
  BYTE  B;
  BYTE  G;
  BYTE  R;
  BYTE  A;
} RGBQ;

#define  BGRC  COLORREF

#define  MMPI  25.4
#define  PTPI  72.0

#define  DEFDPI   96
#define  PRNDPI  300

#define  PT2MM(pt)  ((pt) * MMPI / PTPI)

#define  PT2DOTS(pt)  ((int)((pt) * DEFDPI / PTPI + 0.5))

#define  DIB_24BITS     0x00000000
#define  DIB_32BITS     0x00000020
#define  DIB_BOTTOMUP   0x00000000
#define  DIB_TOPDOWN    0x00000080

#define  DIB_24B  DIB_24BITS
#define  DIB_32B  DIB_32BITS
#define  DIB_BUP  DIB_BOTTOMUP
#define  DIB_TDN  DIB_TOPDOWN

typedef struct { BITMAPINFOHEADER *bi;} HDIB_;
typedef HDIB_ * HDIB;

#define  DIB_INFO(bm)         ((bm)->bi)
#define  DIB_WIDTH(bm)        DIB_INFO(bm)->biWidth
#define  DIB_HEIGHT(bm)       DIB_INFO(bm)->biHeight
#define  DIB_BITS(bm)         DIB_INFO(bm)->biBitCount
#define  DIB_XPPM(bm)         DIB_INFO(bm)->biXPelsPerMeter
#define  DIB_YPPM(bm)         DIB_INFO(bm)->biYPelsPerMeter
#define  DIB_COLOR_USED(bm)   DIB_INFO(bm)->biClrUsed
#define  DIB_HEADER_SIZE(bm)  DIB_INFO(bm)->biSize
#define  DIB_BMI(bm)          ((BITMAPINFO *)DIB_INFO(bm))
#define  DIB_PALETTE(bm)      ((PBYTE)DIB_INFO(bm) + DIB_HEADER_SIZE(bm))

UINT WINAPI DIB_PALETTE_SIZE(HDIB bm);

#define  DIB_IMAGE_OFFSET(bm) (DIB_HEADER_SIZE(bm) + DIB_PALETTE_SIZE(bm))
#define  DIB_IMAGE(bm)        ((PBYTE)DIB_INFO(bm) + DIB_IMAGE_OFFSET(bm))
#define  DIB_LINE_SIZE(bm)    (((DIB_WIDTH(bm) * DIB_BITS(bm) + 31) / 32) * 4)
#define  DIB_IMAGE_SIZE(bm)   (DIB_LINE_SIZE(bm) * abs(DIB_HEIGHT(bm)))
#define  DIB_SIZE(bm)         (DIB_IMAGE_OFFSET(bm) + DIB_IMAGE_SIZE(bm))

#define  DIB_72DPI    2835
#define  DIB_96DPI    3780
#define  DIB_300DPI  11811

#define  DIB_72D  DIB_72DPI
#define  DIB_96D  DIB_96DPI
#define  DIB_300  DIB_300DPI

HDIB WINAPI CreateDIBX(int w, int h, int flg, int xppm, int yppm);
HDIB WINAPI CreateDIB(int w, int h, int flg);
HDIB WINAPI CreateCompatibleDIB(HDIB bm);
HDIB WINAPI DuplicateDIB(HDIB bm);
void WINAPI DeleteDIB(HDIB bm);
HDIB WINAPI MoveDIB(HDIB bm, HDIB bx);

//typedef struct {
//  int   cx;
//  int   cy;
//} SIZE;
//typedef SIZE * PSIZE;

PSIZE WINAPI GetDIBSize(PSIZE sz, HDIB bm);

HDIB WINAPI LoadDIB(HDIB bm, PCSTR nam);
BOOL WINAPI SaveDIB(HDIB bm, PCSTR nam);

// ---------------------------------------------------------------------- //

typedef BYTE (CALLBACK * FC1B)(RGBC cr);

HDIB WINAPI CreateDIBHistgram(HDIB bx, FC1B fext);

typedef BYTE (CALLBACK * FC2B)(RGBC c1, RGBC c2);

BOOL WINAPI QuantizeDIB(HDIB bm, FC2B fdst, int nc);

typedef BYTE (CALLBACK * FCHC)(BYTE i, UINT pa);

void WINAPI SetCharacteristicCurve(BYTE ct[256], FCHC fchr, UINT pa);

typedef RGBC (CALLBACK * FC1C)(RGBC cr);

RGBC WINAPI BGR2RGB(BGRC cr);
BGRC WINAPI RGB2BGR(RGBC cr);
BGRC WINAPI BKG2TXT(BGRC cr);

typedef RGBC (CALLBACK * FC1P)(RGBC cr, UINT pa);

BOOL WINAPI ChangeDIBColor(HDIB bm, FC1P fcvt, UINT pa);

typedef RGBC (CALLBACK * FC2C)(RGBC c1, RGBC c2);

typedef RGBC (CALLBACK * FC2P)(RGBC c1, RGBC c2, UINT pa);

BOOL WINAPI SuperposeDIB(HDIB bm, HDIB bx, FC2P fopr, UINT pa);

typedef RGBC (CALLBACK * FXYP)(int x, int y, UINT pa);

BOOL WINAPI PaintDIB(HDIB bm, FXYP fpnt, UINT pa);

typedef struct {
  RGBC  *wk;
  int   ni;
} WINS;
typedef WINS * PWINS;
typedef const WINS * PCWINS;

typedef PCWINS (CALLBACK * FWIN)(const RGBC *qd, int x, int y, int w, int h);
typedef RGBC (CALLBACK * FFIL)(PCWINS ws, UINT pa);

BOOL WINAPI FilterDIB(HDIB bm, FWIN fwin, FFIL ffil, UINT pa);

BOOL WINAPI TrimDIB(HDIB bm, int x, int y, int w, int h, RGBC cb);
BOOL WINAPI PasteDIB(HDIB bm, HDIB bx, int x, int y);

BOOL WINAPI ResizeDIBA(HDIB bm, int w, int h);
BOOL WINAPI ResizeDIBB(HDIB bm, int w, int h);

#define  DIB_MEAN    0
#define  DIB_LINEAR  1
#define  DIB_SMOOTH  2
#define  DIB_SHARP   3

#define  DIB_MEA  DIB_MEAN
#define  DIB_LIN  DIB_LINEAR
#define  DIB_SMO  DIB_SMOOTH
#define  DIB_SHA  DIB_SHARP

BOOL WINAPI ResizeDIBX(HDIB bm, int w, int h, int sw);

#define  ResizeDIB(bm, w, h)  ResizeDIBX(bm, w, h, DIB_MEA)

#define  DIB_HORIZONTAL  1
#define  DIB_VERTICAL    2

#define  DIB_HRZ  DIB_HORIZONTAL
#define  DIB_VRT  DIB_VERTICAL

BOOL WINAPI InvertDIB(HDIB bm, int sw);
BOOL WINAPI RotateDIB(HDIB bm, REAL dg, RGBC cb);

typedef struct {
  int  x1, y1;
  int  x2, y2;
  int  x3, y3;
  int  x4, y4;
} XFMQ;
typedef XFMQ * PXFMQ;
typedef const XFMQ * PCXFMQ;

void WINAPI SetXFMQ(PXFMQ q, int x1, int y1, int x2, int y2, int x3, int y3, int x4, int y4);

BOOL WINAPI TransformDIB(HDIB bm, PCXFMQ q, RGBC cb);

typedef void (CALLBACK * FXFM)(REAL *px, REAL *py, REAL ix, REAL iy, UINT pa);

BOOL WINAPI TransformDIBX(HDIB bm, int w, int h, FXFM fxfm, UINT pa, RGBC cb);

BOOL WINAPI ConvertDIB(HDIB bm, int flg);

HDIB WINAPI LoadBMP(HDIB bm, PCSTR nam);
BOOL WINAPI SaveBMP(HDIB bm, PCSTR nam);

HDIB WINAPI CaptureDIB(HDIB bm, HDC hdc, PCRECT pr);

HDIB WINAPI CreateDIBFromDDB(HBITMAP hbmp);
HBITMAP WINAPI CreateDDBFromDIB(HDIB bm);

#define  DT_FORCETOP      0x00000004
//#define  DT_LEFT          0x00000000
//#define  DT_CENTER        0x00000001
//#define  DT_RIGHT         0x00000002
#define  DT_DEFAULT       0x00000000
//#define  DT_BOTTOM        0x00000008
//#define  DT_CALCRECT      0x00000400
#define  DT_ADJUSTWIDTH   0x02000000
#define  DT_ADJUSTHEIGHT  0x04000000
#define  DT_GRAYED        0x08000000
#define  DT_STRETCH      (DT_ADJUSTWIDTH|DT_ADJUSTHEIGHT)
#define  DT_CLIP          0x40000000

#define  DT_FRC  DT_FORCETOP
#define  DT_LFT  DT_LEFT
#define  DT_CEN  DT_CENTER
#define  DT_RGT  DT_RIGHT
#define  DT_MID  DT_DEFAULT
#define  DT_BTM  DT_BOTTOM
#define  DT_CAL  DT_CALCRECT
#define  DT_AJW  DT_ADJUSTWIDTH
#define  DT_AJH  DT_ADJUSTHEIGHT
#define  DT_GRY  DT_GRAYED
#define  DT_STB  DT_STRETCH
#define  DT_CLP  DT_CLIP

#define  DTisMID(fmt)  (!(fmt & (DT_FRC|DT_BTM)))

BOOL WINAPI DrawDIB(HDC hdc, PRECT pr, HDIB bm, int fmt);
BOOL WINAPI DrawDDB(HDC hdc, PRECT pr, HBITMAP hbmp, int fmt);

// ---------------------------------------------------------------------- //

int  WINAPI NAppColors(void);
void WINAPI SetAppColor(int cid, BGRC col);
BGRC WINAPI AppColor(int cid);
char WINAPI AppColorChar(int cid);
int  WINAPI AppColorIndex(char ch);
int  WINAPI NearestAppColorIndex(BGRC col);

HDIB WINAPI CreateDIBFromSTR(int w, int h, const char bi[]);
HBITMAP WINAPI CreateDDBFromSTR(int w, int h, const char bi[]);

#define  CID_WINDOWTEXT     0
#define  CID_RED            2
#define  CID_WINDOW         7
#define  CID_3DSHADOW       8
#define  CID_3DFACE        15
#define  CID_3DDKSHADOW    16
#define  CID_3DHILIGHT     23
#define  CID_CELLPTR       24
#define  CID_INDEXPTR      31
#define  CID_HIGHLIGHT     32
#define  CID_HIGHLIGHTTEXT 40
#define  CID_3DLIGHT       48
#define  CID_BLACK         49
#define  CID_GRAY          55
#define  CID_WHITE         63

#define  CID_TXT  CID_WINDOWTEXT
#define  CID_WIN  CID_WINDOW
#define  CID_3DS  CID_3DSHADOW
#define  CID_3DF  CID_3DFACE
#define  CID_3DD  CID_3DDKSHADOW
#define  CID_3DH  CID_3DHILIGHT
#define  CID_PTR  CID_CELLPTR
#define  CID_IDX  CID_INDEXPTR
#define  CID_HIL  CID_HIGHLIGHT
#define  CID_HIT  CID_HIGHLIGHTTEXT
#define  CID_3DL  CID_3DLIGHT
#define  CID_BLK  CID_BLACK
#define  CID_GRY  CID_GRAY
#define  CID_WHI  CID_WHITE

HPEN WINAPI AppPen(int cid);

HBRUSH WINAPI AppBrush(int cid);

#define  TFF_BOLD       0x0070
#define  TFF_ITALIC     0x0100
#define  TFF_UNDERLINE  0x0200
#define  TFF_STRIKEOUT  0x0400
#define  TFF_VERTICAL   0x1000

#define  TFF_BLD  TFF_BOLD
#define  TFF_ITA  TFF_ITALIC
#define  TFF_UND  TFF_UNDERLINE
#define  TFF_STO  TFF_STRIKEOUT
#define  TFF_VRT  TFF_VERTICAL

PLOGFONT WINAPI SetLOGFONT(PLOGFONT lf, PCSTR face, int ch, int flg);

void WINAPI GetTextSize(PSIZE sz, HDC hdc, PCLOGFONT lf, PCSTR st, int cb);
void WINAPI GetAveCharSize(PSIZE sz, HDC hdc, PCLOGFONT lf);
void WINAPI GetZenCharSize(PSIZE sz, HDC hdc, PCLOGFONT lf);

int  WINAPI NAppFonts(void);
void WINAPI SetAppFont(int fid, HFONT hfnt);

#define  FID_DIALOG       0
#define  FID_SYSTEM       1
#define  FID_DEFAULT      2
#define  FID_MONOSPACE    3
#define  FID_WEBDINGS     4
#define  FID_WINGDINGS    5
#define  FID_USER         6

#define  FID_DLG  FID_DIALOG
#define  FID_SYS  FID_SYSTEM
#define  FID_DEF  FID_DEFAULT
#define  FID_MON  FID_MONOSPACE
#define  FID_WEB  FID_WEBDINGS
#define  FID_WIN  FID_WINGDINGS
#define  FID_USR  FID_USER

HFONT WINAPI AppFont(int fid);

int  WINAPI NAppDDBs(void);
void WINAPI SetAppDDB(int bid, HBITMAP hbmp);

#define  MNUBMPW  16
#define  MNUBMPH  16

#define  BID_BLANK    0
#define  BID_NEW      1
#define  BID_OPEN     2
#define  BID_CLOSE    3
#define  BID_SAVE     4
#define  BID_SAVEAS   5
#define  BID_PRINT    6
#define  BID_SETUP    7
#define  BID_CUT      8
#define  BID_COPY     9
#define  BID_PASTE   10
#define  BID_SEARCH  11
#define  BID_SORT    12
#define  BID_FILTER  13
#define  BID_SUMMARY 14
#define  BID_CHECK   15
#define  BID_PREV    16
#define  BID_NEXT    17
#define  BID_INVERT  18
#define  BID_ZOOMIN  19
#define  BID_ZOOMOUT 20
#define  BID_SAME    21
#define  BID_WHOLE   22
#define  BID_PREVWIN 23
#define  BID_NEXTWIN 24
#define  BID_VIEW    25
#define  BID_ARRANGE 26
#define  BID_EDIT    27
#define  BID_ADD     28
#define  BID_INSERT  29
#define  BID_DELETE  30
#define  BID_COMPILE 31

#define  MNDBMPW  32
#define  MNDBMPH  32

#define  BID_DUMMY   48
#define  BID_INPUT   49
#define  BID_TABLE   50
#define  BID_EXECUTE 51
#define  BID_REPORT  52
#define  BID_TEXT    53
#define  BID_SYSTEM  54
#define  BID_EXIT    55
#define  BID_BACK    56

HBITMAP WINAPI AppDDB(int bid);

void WINAPI SetMenuItemDDB(int idm, HBITMAP hbmp);

HBITMAP WINAPI MenuItemDDB(int idm);

#define  DLGFACE  "MS UI Gothic"
#define  DLGPTSZ  9

#define  DEFFACE  "lr oSVbN"
#define  DEFPTSZ  11

#define  MONFACE  "lr SVbN"
#define  MONPTSZ  11

// ---------------------------------------------------------------------- //

typedef struct { int unused;} HCVS_;
typedef HCVS_ * HCVS;

HCVS WINAPI CreateCanvas(HDC hdc, UINT xsz);
void WINAPI DeleteCanvas(HCVS cv);

HDC  WINAPI SetCanvasHDC(HCVS cv, HDC hdc, int dpi);
HDC  WINAPI CanvasHDC(HCVS cv);
HDC  WINAPI CanvasHIC(HCVS cv);
BOOL WINAPI EnableCanvas(HCVS cv);
BOOL WINAPI DisableCanvas(HCVS cv);
PLOGFONT WINAPI CanvasLOGFONT(PLOGFONT lf, HCVS cv);

void WINAPI SetOrigin(HCVS cv, REAL xorg, REAL yorg);
REAL WINAPI OriginX(HCVS cv);
REAL WINAPI OriginY(HCVS cv);
void WINAPI SetScale(HCVS cv, REAL xscl, REAL yscl);
REAL WINAPI ScaleX(HCVS cv);
REAL WINAPI ScaleY(HCVS cv);

int  WINAPI c2dW(HCVS cv, REAL w);
int  WINAPI c2dH(HCVS cv, REAL h);
int  WINAPI c2dX(HCVS cv, REAL x);
int  WINAPI c2dY(HCVS cv, REAL y);
REAL WINAPI d2cW(HCVS cv, int w);
REAL WINAPI d2cH(HCVS cv, int h);
REAL WINAPI d2cX(HCVS cv, int x);
REAL WINAPI d2cY(HCVS cv, int y);

typedef struct {
  REAL  x1;
  REAL  y1;
  REAL  x2;
  REAL  y2;
} CELL;
typedef CELL * PCELL;
typedef const CELL * PCCELL;

PCELL WINAPI SetCell(PCELL pc, REAL x1, REAL y1, REAL x2, REAL y2);

PRECT WINAPI c2dRect(PRECT pr, HCVS cv, PCCELL pc);
PCELL WINAPI d2cCell(PCELL pc, HCVS cv, PCRECT pr);

PCELL WINAPI MoveCell(PCELL pc, REAL dx, REAL dy);
PCELL WINAPI ResizeCell(PCELL pc, REAL dx, REAL dy);
PCELL WINAPI ReshapeCell(PCELL pc, REAL lm, REAL tm, REAL rm, REAL bm);

void WINAPI SetCellMargin(HCVS cv, REAL hcmg, REAL vcmg);
#define  CellMarginHD(cv)  c2dW(cv, CellMarginH(cv))
REAL WINAPI CellMarginH(HCVS cv);
#define  CellMarginVD(cv)  c2dH(cv, CellMarginV(cv))
REAL WINAPI CellMarginV(HCVS cv);

#define  LST_NULL        0x00
#define  LST_THIN        0x01
#define  LST_REGULAR     0x02
#define  LST_MEDIUM      0x03
#define  LST_THICK       0x04
#define  LST_EXTRATHICK  0x05
#define  LST_DASH        0x10
#define  LST_DOT         0x20
#define  LST_DASHDOT     0x30
#define  LST_DASHDOTDOT  0x40

#define  LST_NUL  LST_NULL
#define  LST_THN  LST_THIN
#define  LST_REG  LST_REGULAR
#define  LST_MED  LST_MEDIUM
#define  LST_THK  LST_THICK
#define  LST_XTK  LST_EXTRATHICK
#define  LST_DSH  LST_DASH
#define  LST_DSD  LST_DASHDOT
#define  LST_DDD  LST_DASHDOTDOT

#define  lstN  LST_NUL
#define  lstI  LST_THN
#define  lstR  LST_REG
#define  lstM  LST_MED
#define  lstB  LST_THK
#define  lstX  LST_XTK
#define  lstD  LST_DOT

int  WINAPI SetLineStyle(HCVS cv, int lst);
BGRC WINAPI SetLineColor(HCVS cv, BGRC col);

#define  SetTextFontD(cv, face, ch, flg, col)  SetTextFont(cv, face, d2cH(cv, ch), flg, col)
void WINAPI SetTextFont(HCVS cv, PCSTR face, REAL ch, int flg, BGRC col);
#define  SetTextStyleD(cv, cw, ch, cd, face)  SetTextStyle(cv, d2cW(cv, cw), d2cH(cv, ch), d2cW(cv, cd), face)
void WINAPI SetTextStyle(HCVS cv, REAL cw, REAL ch, REAL cd, PCSTR face);

typedef struct {
  REAL  w;
  REAL  h;
} WHSZ;
typedef WHSZ * PWHSZ;

PWHSZ WINAPI TextSize(PWHSZ sz, HCVS cv, PCSTR st, int cb);
PSIZE WINAPI TextSizeD(PSIZE sz, HCVS cv, PCSTR st, int cb);
PWHSZ WINAPI AveCharSize(PWHSZ sz, HCVS cv);
PSIZE WINAPI AveCharSizeD(PSIZE sz, HCVS cv);
PWHSZ WINAPI ZenCharSize(PWHSZ sz, HCVS cv);
PSIZE WINAPI ZenCharSizeD(PSIZE sz, HCVS cv);

void WINAPI DrawLineD(HCVS cv, int x1, int y1, int x2, int y2);
void WINAPI DrawLine(HCVS cv, REAL x1, REAL y1, REAL x2, REAL y2);
#define  DrawCircD(cv, xc, yc, rc, t1, t2)  DrawCirc(cv, d2cX(cv, xc), d2cY(cv, yc), d2cW(cv, rc), t1, t2)
void WINAPI DrawCirc(HCVS cv, REAL xc, REAL yc, REAL rc, REAL t1, REAL t2);

#define  RST_NULL                 0
#define  RST_DEFAULT            (-1)
#define  RST_THIN        0x01010101
#define  RST_REGULAR     0x02020202
#define  RST_MEDIUM      0x03030303
#define  RST_THICK       0x04040404
#define  RST_EXTRATHICK  0x05050505
#define  RST_DASH        0x10101010
#define  RST_DOT         0x20202020
#define  RST_DASHDOT     0x30303030
#define  RST_DASHDOTDOT  0x40404040

#define  RST_NUL  RST_NULL
#define  RST_DEF  RST_DEFAULT
#define  RST_THN  RST_THIN
#define  RST_REG  RST_REGULAR
#define  RST_MED  RST_MEDIUM
#define  RST_THK  RST_THICK
#define  RST_XTK  RST_EXTRATHICK
#define  RST_DSH  RST_DASH
#define  RST_DSD  RST_DASHDOT
#define  RST_DDD  RST_DASHDOTDOT

#define  RST(llst, tlst, rlst, blst)  MAKEUINT(MAKEWORD(llst, tlst), MAKEWORD(rlst, blst))

void WINAPI DrawCellD(HCVS cv, PCRECT pr, int rst);
void WINAPI DrawCell(HCVS cv, PCCELL pc, int rst);
void WINAPI FillCellD(HCVS cv, PCRECT pr, int rst, BGRC col);
void WINAPI FillCell(HCVS cv, PCCELL pc, int rst, BGRC col);

typedef struct {
  REAL  x;
  REAL  y;
} XYPT;
typedef XYPT * PXYPT;

void WINAPI FillPolyD(HCVS cv, POINT pt[], int np, BGRC col);
void WINAPI FillPoly(HCVS cv, XYPT pt[], int np, BGRC col);
#define  FillCircD(cv, xc, yc, rc, t1, t2, col)  FillCirc(cv, d2cX(cv, xc), d2cY(cv, yc), d2cW(cv, rc), t1, t2, col)
void WINAPI FillCirc(HCVS cv, REAL xc, REAL yc, REAL rc, REAL t1, REAL t2, BGRC col);

#define  DT_FORCETOP      0x00000004
//#define  DT_LEFT          0x00000000
//#define  DT_CENTER        0x00000001
//#define  DT_RIGHT         0x00000002
#define  DT_DEFAULT       0x00000000
//#define  DT_BOTTOM        0x00000008
//#define  DT_CALCRECT      0x00000400
//#define  DT_PATH_ELLIPSIS 0x00004000
//#define  DT_END_ELLIPSIS  0x00008000
#define  DT_MULTILINE     0x01000000
#define  DT_JUSTIFY       0x02000000
#define  DT_AUTONARROWING 0x04000000
#define  DT_PREFIX        0x08000000
#define  DT_CLIP          0x40000000

#define  DT_FRC  DT_FORCETOP
#define  DT_LFT  DT_LEFT
#define  DT_CEN  DT_CENTER
#define  DT_RGT  DT_RIGHT
#define  DT_MID  DT_DEFAULT
#define  DT_BTM  DT_BOTTOM
#define  DT_CAL  DT_CALCRECT
#define  DT_PEL  DT_PATH_ELLIPSIS
#define  DT_EEL  DT_END_ELLIPSIS
#define  DT_MUL  DT_MULTILINE
#define  DT_JST  DT_JUSTIFY
#define  DT_AUT  DT_AUTONARROWING
#define  DT_PRE  DT_PREFIX
#define  DT_CLP  DT_CLIP

void WINAPI CellTextD(HCVS cv, PRECT pc, PCSTR st, int fmt);
void WINAPI CellText(HCVS cv, PCELL pc, PCSTR st, int fmt);

#define  VT_LEFT          0x00000001
#define  VT_CENTER        0x00000000
#define  VT_RIGHT         0x00000002
#define  VT_TOP           0x00000000
#define  VT_VCENTER       0x00000004
#define  VT_BOTTOM        0x00000008
#define  VT_CALCRECT      0x00000400

#define  VT_LFT  VT_LEFT
#define  VT_CEN  VT_CENTER
#define  VT_RGT  VT_RIGHT
#define  VT_MID  VT_VCENTER
#define  VT_BTM  VT_BOTTOM
#define  VT_CAL  VT_CALCRECT

void WINAPI VCellTextD(HCVS cv, PRECT pc, PCSTR st, int fmt);
void WINAPI VCellText(HCVS cv, PCELL pc, PCSTR st, int fmt);

typedef struct {
  REAL  x0;
  REAL  y0;
  REAL  dcw;
  REAL  drh;
  REAL  *cw;
  int   nc;
  REAL  *rh;
  int   nr;
} GRID;
typedef GRID * PGRID;
typedef const GRID * PCGRID;

void WINAPI SetGrid0(PGRID gr, REAL x0, REAL y0, int nc, int nr, REAL dcw, REAL drh);
void WINAPI SetGrid1(PGRID gr, REAL x0, REAL y0, int nc, int nr, REAL *cw, REAL drh);
void WINAPI SetGrid2(PGRID gr, REAL x0, REAL y0, int nc, int nr, REAL dcw, REAL *rh);
void WINAPI SetGrid3(PGRID gr, REAL x0, REAL y0, int nc, int nr, REAL *cw, REAL *rh);

PGRID WINAPI AdjustGrid(HCVS cv, PGRID gr);
PCELL WINAPI GridCell(PCELL pc, PCGRID gr, int ir, int ic);

#define  RST_THIN_REGULAR       (-2)
#define  RST_THIN_MEDIUM        (-3)
#define  RST_THIN_THICK         (-4)
#define  RST_REGULAR_MEDIUM     (-5)
#define  RST_REGULAR_THICK      (-6)
#define  RST_REGULAR_EXTRATHICK (-7)

#define  RST_I_R  RST_THIN_REGULAR
#define  RST_I_M  RST_THIN_MEDIUM
#define  RST_I_B  RST_THIN_THICK
#define  RST_R_M  RST_REGULAR_MEDIUM
#define  RST_R_B  RST_REGULAR_THICK
#define  RST_R_X  RST_REGULAR_EXTRATHICK

void WINAPI DrawGridCell(HCVS cv, PCGRID gr, int ir, int ic, int rst);
void WINAPI GridCellText(HCVS cv, PCGRID gr, int ir, int ic, int rst, PCSTR st, int fmt);

typedef struct {
  int   sty;
  PCSTR st1;
  PCSTR st2;
  PCSTR st3;
  REAL  dc;
} STMP;
typedef STMP * PSTMP;
typedef const STMP * PCSTMP;

void WINAPI SetStamp0(PSTMP ps, REAL dc, PCSTR st);
void WINAPI SetStamp1(PSTMP ps, REAL dc, PCSTR st);
void WINAPI SetStamp2(PSTMP ps, REAL dc, PCSTR st1, PCSTR st2);
void WINAPI SetStamp3(PSTMP ps, REAL dc, PCSTR st1, PCSTR st2, PCSTR st3);

void WINAPI CellStampD(HCVS cv, PCRECT pr, PCSTMP ps);
void WINAPI CellStamp(HCVS cv, PCCELL pc, PCSTMP ps);

void WINAPI CellImageD(HCVS cv, PRECT pc, HDIB bm, int fmt);
void WINAPI CellImage(HCVS cv, PCELL pc, HDIB bm, int fmt);

// ---------------------------------------------------------------------- //

//#define  WS_OVERLAPPED       0x00000000
//#define  WS_POPUP            0x80000000
//#define  WS_CHILD            0x40000000
//#define  WS_MINIMIZE         0x20000000
//#define  WS_VISIBLE          0x10000000
//#define  WS_DISABLED         0x08000000
//#define  WS_CLIPSIBLINGS     0x04000000
//#define  WS_CLIPCHILDREN     0x02000000
//#define  WS_MAXIMIZE         0x01000000
//#define  WS_CAPTION          0x00C00000
//#define  WS_BORDER           0x00800000
//#define  WS_DLGFRAME         0x00400000
//#define  WS_VSCROLL          0x00200000
//#define  WS_HSCROLL          0x00100000
//#define  WS_SYSMENU          0x00080000
//#define  WS_THICKFRAME       0x00040000
//#define  WS_GROUP            0x00020000
//#define  WS_TABSTOP          0x00010000
//#define  WS_MINIMIZEBOX      0x00020000
//#define  WS_MAXIMIZEBOX      0x00010000
//#define  WS_OVERLAPPEDWINDOW 0x00CF0000
//#define  WS_POPUPWINDOW      0x80880000

//#define  DS_MODALFRAME       0x0080

#define  ws_    0
#define  wsT    WS_TABSTOP
#define  wsG    WS_GROUP
#define  wsGT   (WS_GROUP|WS_TABSTOP)
#define  wsH    WS_HSCROLL
#define  wsV    WS_VSCROLL
#define  wsHV   (WS_HSCROLL|WS_VSCROLL)
#define  wsB    WS_BORDER
#define  wsBV   (WS_BORDER|WS_VSCROLL)
#define  wsBH   (WS_BORDER|WS_HSCROLL)
#define  wsBHV  (WS_BORDER|WS_HSCROLL|WS_VSCROLL)
#define  wsC    (WS_CHILD|WS_VISIBLE)
#define  wsCV   (WS_CHILD|WS_VISIBLE|WS_VSCROLL)
#define  wsCH   (WS_CHILD|WS_VISIBLE|WS_HSCROLL)
#define  wsCHV  (WS_CHILD|WS_VISIBLE|WS_HSCROLL|WS_VSCROLL)
#define  wsCB   (WS_CHILD|WS_VISIBLE|WS_BORDER)
#define  wsCBV  (WS_CHILD|WS_VISIBLE|WS_BORDER|WS_VSCROLL)
#define  wsCBH  (WS_CHILD|WS_VISIBLE|WS_BORDER|WS_HSCROLL)
#define  wsCBHV (WS_CHILD|WS_VISIBLE|WS_BORDER|WS_HSCROLL|WS_VSCROLL)
#define  wsD    WS_DISABLED
#define  wsDlg  (WS_POPUP|WS_CAPTION|WS_SYSMENU|DS_MODALFRAME)
#define  wsMls  (wsDlg|WS_VISIBLE)

#define  WindowInstance(hwnd)       (HINSTANCE)GetWindowLong(hwnd, GWL_HINSTANCE)
#define  WindowID(hwnd)             GetWindowLong(hwnd, GWL_ID)
#define  WindowStyle(hwnd)          (UINT)GetWindowLong(hwnd, GWL_STYLE)
#define  WindowExStyle(hwnd)        (UINT)GetWindowLong(hwnd, GWL_EXSTYLE)

#define  SubclassWindow(hwnd, lpfn)    (WNDPROC)SetWindowLong(hwnd, GWL_WNDPROC, (LPARAM)(lpfn))
#define  ResizeWindow(hwnd, cx, cy)    SetWindowPos(hwnd, NULL, 0, 0, cx, cy, SWP_NOMOVE|SWP_NOZORDER)
#define  EnableDlgItem(hdlg, id, sw)   EnableWindow(GetDlgItem(hdlg, id), sw)
#define  ParentNotify(hwnd, nc, lpa)   SendMessage(GetParent(hwnd), WM_COMMAND, MAKEUINT(GetDlgCtrlID(hwnd), nc), (LPARAM)(lpa))

PCSTR WINAPI WindowText(HWND hwnd);
PCSTR WINAPI ClassName(HWND hwnd);

void WINAPI CenterWindow(HWND hwnd, HWND hown);
void WINAPI AdjustWindow(HWND hwnd, PCRECT pr);

int  WINAPI SetHScrollBar(HWND hwnd, int min, int max, int psz, int pos);
int  WINAPI SetVScrollBar(HWND hwnd, int min, int max, int psz, int pos);

// ---------------------------------------------------------------------- //

//#define  MF_ENABLED          0x00000000
//#define  MF_GRAYED           0x00000001
//#define  MF_CHECKED          0x00000008
//#define  MF_POPUP            0x00000010
//#define  MF_END              0x00000080

#define  mf_    MF_ENABLED
#define  mfG    MF_GRAYED
#define  mfC    MF_CHECKED
#define  mfP    MF_POPUP
#define  mfE    MF_END
#define  mfPE   (MF_POPUP|MF_END)
#define  mfGE   (MF_GRAYED|MF_END)

typedef struct {
  int   flg;
  int   idm;
  PCSTR txt;
  UINT  rsv;
} MNUI;
typedef MNUI * PMNUI;
typedef const MNUI * PCMNUI;

typedef struct {
  PMNUI mi;
  int   ni;
} MNUS;
typedef MNUS * PMNUS;
typedef const MNUS * PCMNUS;

HMENU WINAPI CreateMenuBar(PCMNUS ms);
BOOL WINAPI ChangeMenuBar(HMENU hmnu, int pos, PCMNUS ms);

void WINAPI MenuSlice(PINT p1, PINT p2, PCMNUS ms, int pos);

// ---------------------------------------------------------------------- //

typedef struct {
  int   x, y;
  PCSTR txt;
} TIPI;
typedef TIPI * PTIPI;
typedef const TIPI * PCTIPI;

HWND WINAPI CreateToolTip(HWND hwnd, PCTIPI pi);

#define  TOBM_GETNITEMS        (WM_USER + 0x01)
#define  TOBM_GETITEMINFO      (WM_USER + 0x02)
#define  TOBM_RESETITEMS       (WM_USER + 0x03)

#define  TOBI_SPACE   0
#define  TOBI_VLINE   1
#define  TOBI_BUTTON  2
#define  TOBI_SLEFT   3
#define  TOBI_SCENTER 4
#define  TOBI_SRIGHT  5

#define  tbfV  TOBI_VLINE
#define  tbfB  TOBI_BUTTON

#define  tbfS(w)  MAKEUINT(TOBI_SPACE, w)
#define  tbfL(w)  MAKEUINT(TOBI_SLEFT, w)
#define  tbfC(w)  MAKEUINT(TOBI_SCENTER, w)
#define  tbfR(w)  MAKEUINT(TOBI_SRIGHT, w)

#define  TOBWINH  28
#define  STBWINH  22
#define  TOBSEPW   6
#define  TOBMGNW   2
#define  TOBBTNW  24
#define  TOBBTNH  24
#define  TOBSTAH  16

typedef struct {
  int   flg;
  int   idm;
  PCSTR txt;
  UINT  rsv;
} TOBI;
typedef TOBI * PTOBI;
typedef const TOBI * PCTOBI;

typedef struct {
  PTOBI pi;
  int   ni;
} TOBS;
typedef TOBS * PTOBS;
typedef const TOBS * PCTOBS;

HWND WINAPI CreateToolBar(HWND hwnd, int id, PCTOBS ts);
HWND WINAPI CreateStatBar(HWND hwnd, int id, PCTOBS ss);

// ---------------------------------------------------------------------- //

#define  SetSSTR(hdlg, id, st)  SetDlgItemText(hdlg, id, st)
#define  GetSSTR(hdlg, id, st, len)  GetDlgItemText(hdlg, id, st, len)

#define  SetBSTR(hdlg, id, st)  SetDlgItemText(hdlg, id, st)
#define  GetBSTR(hdlg, id, st, len)  GetDlgItemText(hdlg, id, st, len)

//#define  BST_UNCHECKED       0x0000
//#define  BST_CHECKED         0x0001

#define  SetBFLG(hdlg, id, va) \
  CheckDlgButton(hdlg, id, (va)? BST_CHECKED: BST_UNCHECKED)
#define  GetBFLG(hdlg, id) \
  (BOOL)(IsDlgButtonChecked(hdlg, id) == BST_CHECKED)

#define  SetBSEL(hdlg, id, va) \
  CheckDlgButton(hdlg, id, (va)? BST_CHECKED: BST_UNCHECKED)
#define  GetBSEL(hdlg, id) \
  (BOOL)(IsDlgButtonChecked(hdlg, id) == BST_CHECKED)

#define  SetESTR(hdlg, id, st)  SetDlgItemText(hdlg, id, st)
#define  GetESTR(hdlg, id, st, len)  GetDlgItemText(hdlg, id, st, len)

void WINAPI SetEINT(HWND hdlg, int id, int va);
int  WINAPI GetEINT(HWND hdlg, int id);

void WINAPI SetEDBL(HWND hdlg, int id, REAL va);
REAL WINAPI GetEDBL(HWND hdlg, int id);

void WINAPI SetLSEL(HWND hdlg, int id, int ir, PCENMS ls);
int  WINAPI GetLSEL(HWND hdlg, int id);

void WINAPI SetCSEL(HWND hdlg, int id, int ir, PCENMS ls);
int  WINAPI GetCSEL(HWND hdlg, int id);

void WINAPI SetCCOD(HWND hdlg, int id, int va, PCCODS ls);
int  WINAPI GetCCOD(HWND hdlg, int id, PCCODS ls);

void WINAPI SetCSTR(HWND hdlg, int id, PCSTR st, PCENMS ls);
#define  GetCSTR(hdlg, id, st, len)  GetDlgItemText(hdlg, id, st, len)

//#define  SS_LEFT             0x0000
//#define  SS_CENTER           0x0001
//#define  SS_RIGHT            0x0002
//#define  SS_ICON             0x0003
//#define  SS_OWNERDRAW        0x000d
//#define  SS_BITMAP           0x000e
//#define  SS_CENTERIMAGE      0x0200

#define  ssL    (SS_CENTERIMAGE|SS_LEFT)
#define  ssC    (SS_CENTERIMAGE|SS_CENTER)
#define  ssR    (SS_CENTERIMAGE|SS_RIGHT)
#define  ssM    SS_LEFT
#define  ssI    SS_ICON
#define  ssO    SS_OWNERDRAW
#define  ssB    SS_BITMAP

//#define  BS_PUSHBUTTON       0x0000
//#define  BS_DEFPUSHBUTTON    0x0001
//#define  BS_AUTOCHECKBOX     0x0003
//#define  BS_GROUPBOX         0x0007
//#define  BS_AUTORADIOBUTTON  0x0009
//#define  BS_OWNERDRAW        0x000b
//#define  BS_FLAT             0x8000

#define  bsP    BS_PUSHBUTTON
#define  bsD    BS_DEFPUSHBUTTON
#define  bsC    BS_AUTOCHECKBOX
#define  bsR    BS_AUTORADIOBUTTON
#define  bsG    BS_GROUPBOX
#define  bsO    BS_OWNERDRAW
#define  bsF    BS_FLAT

//#define  ES_LEFT             0x0000
//#define  ES_CENTER           0x0001
//#define  ES_RIGHT            0x0002
//#define  ES_MULTILINE        0x0004
//#define  ES_PASSWORD         0x0020
//#define  ES_AUTOVSCROLL      0x0040
//#define  ES_AUTOHSCROLL      0x0080
//#define  ES_READONLY         0x0800
//#define  ES_WANTRETURN       0x1000

#define  esV    ES_AUTOVSCROLL
#define  esH    ES_AUTOHSCROLL
#define  esP    ES_PASSWORD
#define  esC    ES_CENTER
#define  esR    ES_RIGHT
#define  esM    (ES_MULTILINE|ES_WANTRETURN)

//#define  LBS_NOTIFY          0x0001
//#define  LBS_USETABSTOPS     0x0080

#define  lbsNT  (LBS_NOTIFY|LBS_USETABSTOPS)

//#define  CBS_DROPDOWN        0x0002
//#define  CBS_DROPDOWNLIST    0x0003
//#define  CBS_AUTOHSCROLL     0x0040

#define  cbsL   CBS_DROPDOWNLIST
#define  cbsDH  (CBS_DROPDOWN|CBS_AUTOHSCROLL)

//#define  PSH_PROPSHEETPAGE   0x0008
//#define  PSH_WIZARD          0x0020
//#define  PSH_NOAPPLYNOW      0x0080

#define  pshP   PSH_PROPSHEETPAGE
#define  pshW   PSH_WIZARD
#define  pshN   PSH_NOAPPLYNOW

#define  C_SLFT  "static", wsC | ssL
#define  C_SCEN  "static", wsC | ssC
#define  C_SRGT  "static", wsC | ssR
#define  C_SMUL  "static", wsC | ssM
#define  C_SICO  "static", wsC | ssI
#define  C_SOWN  "static", wsC | ssO
#define  C_SBMP  "static", wsC | ssB
#define  C_BPSH  "button", wsC | bsP
#define  C_BDEF  "button", wsC | bsD
#define  C_BFLG  "button", wsC | bsC
#define  C_BSEL  "button", wsC | bsR
#define  C_BGRP  "button", wsC | bsG
#define  C_BOWN  "button", wsC | bsO
#define  C_BFLT  "button", wsC | bsF
#define  C_ENUM  "edit", wsCB | esR
#define  C_ESTR  "edit", wsCB | esH
#define  C_EPWD  "edit", wsCB | esP
#define  C_EMUL  "edit", wsCBV | esM | esV
#define  C_LSEL  "listbox", wsCBV
#define  C_LTBL  "listbox", wsCBV | lbsNT
#define  C_CSEL  "combobox", wsCBV | cbsL
#define  C_CSTR  "combobox", wsCBV | cbsDH

typedef struct {
  PCSTR cls;
  int   sty;
  WORD  x, y;
  WORD  w, h;
  int   id;
  PCSTR txt;
} DLGI;
typedef DLGI * PDLGI;
typedef const DLGI * PCDLGI;

typedef struct {
  PDLGI di;
  int   ni;
  WORD  x, y;
  WORD  w, h;
  PCSTR tit;
} DLGS;
typedef DLGS * PDLGS;
typedef const DLGS * PCDLGS;

void WINAPI SetDLGI(PDLGI di, PCSTR cls, int sty, WORD x, WORD y, WORD w, WORD h, int id, PCSTR txt);

int  WINAPI DlgBoxInput(HWND hwnd, PCDLGS ds, DLGPROC lpfn, LPARAM lpa);
HWND WINAPI CreateDlgBox(HWND hwnd, PCDLGS ds, DLGPROC lpfn, LPARAM lpa);

#define  IDPSH  900

typedef struct {
  PCDLGS ds;
  DLGPROC lpfn;
  LPARAM lpa;
  UINT   psz;
} PSHI;
typedef PSHI * PPSHI;
typedef const PSHI * PCPSHI;

typedef struct {
  PPSHI pp;
  int   np;
} PSHS;
typedef PSHS * PPSHS;
typedef const PSHS * PCPSHS;

int  WINAPI PropSheetInput(HWND hwnd, PCPSHS ps, int ip);

//#define  IDOK      1
//#define  IDCANCEL  2
//#define  IDABORT   3
//#define  IDRETRY   4
//#define  IDIGNORE  5
//#define  IDYES     6
//#define  IDNO      7

int  WINAPI MsgBox(HWND hwnd, PCSTR st, PCSTR ca, int flg);

int  WINAPI OkBox(HWND hwnd, PCSTR st, PCSTR ca);
int  WINAPI OkCanBox(HWND hwnd, PCSTR st, PCSTR ca);
int  WINAPI YesNoBox(HWND hwnd, PCSTR st, PCSTR ca);
int  WINAPI YesNoCanBox(HWND hwnd, PCSTR st, PCSTR ca);
int  WINAPI ErrorBox(HWND hwnd, PCSTR st, PCSTR ca);
int  WINAPI WarningBox(HWND hwnd, PCSTR st, PCSTR ca);
int  WINAPI AlertBox(HWND hwnd, PCSTR st, PCSTR ca);
int  WINAPI ResultBox(HWND hwnd, PCSTR st, PCSTR ca, BOOL re);


//#define  OFN_READONLY        0x00000001
//#define  OFN_OVERWRITEPROMPT 0x00000002
//#define  OFN_HIDEREADONLY    0x00000004
//#define  OFN_NOCHANGEDIR     0x00000008

typedef struct {
  PCSTR fil;
  int   idx;
  PSTR  nam;
  int   flg;
} OFNS;
typedef OFNS * POFNS;

BOOL WINAPI OpenFileDlgBox(HWND hwnd, POFNS os);
BOOL WINAPI SaveFileDlgBox(HWND hwnd, POFNS os);

//#define  PD_ALLPAGES         0x00000000
//#define  PD_SELECTION        0x00000001
//#define  PD_PAGENUMS         0x00000002
//#define  PD_NOSELECTION      0x00000004
//#define  PD_NOPAGENUMS       0x00000008
//#define  PD_RETURNDC         0x00000100
//#define  PD_RETURNDEFAULT    0x00000400

typedef struct {
  HDC   hdc;
  int   flg;
  WORD  nf, nt;
  WORD  np, nc;
  WORD  psz, ori;
  WORD  wid, len;
} PRDS;
typedef PRDS * PPRDS;

BOOL WINAPI PrintDlgBox(HWND hwnd, PPRDS ps);

typedef struct {
  int   adj;
  int   apw;
  int   opw, oph;
  int   ppw, pph;
  int   lpm, tpm, rpm, bpm;
} PSDS;
typedef PSDS * PPSDS;
typedef const PSDS * PCPSDS;

BOOL WINAPI PageSetupDlgBox(HWND hwnd, PPSDS ps);

// ---------------------------------------------------------------------- //

#define  REPM_SKIPPAGE   0
#define  REPM_DRAWPAGE   1
#define  REPM_STARTPRINT 2
#define  REPM_ENDPRINT   3
#define  REPM_SETUP      4
#define  REPM_PAGESETUP  5
#define  REPM_CREATE     6
#define  REPM_DESTROY    7

typedef HCVS  HREP;

typedef BOOL (CALLBACK * FREP)(HREP rp, int msg, int pg, UINT pa);

#define  PSZ_A3_PORTRAIT   0x00
#define  PSZ_A3_LANDSCAPE  0x08
#define  PSZ_B4_PORTRAIT   0x01
#define  PSZ_B4_LANDSCAPE  0x09
#define  PSZ_A4_PORTRAIT   0x02
#define  PSZ_A4_LANDSCAPE  0x0a
#define  PSZ_B5_PORTRAIT   0x03
#define  PSZ_B5_LANDSCAPE  0x0b
#define  PSZ_A5_PORTRAIT   0x04
#define  PSZ_A5_LANDSCAPE  0x0c
#define  PSZ_B6_PORTRAIT   0x05
#define  PSZ_B6_LANDSCAPE  0x0d
#define  PSZ_A6_PORTRAIT   0x06
#define  PSZ_A6_LANDSCAPE  0x0e

#define  PSZ_A3P  PSZ_A3_PORTRAIT
#define  PSZ_A3L  PSZ_A3_LANDSCAPE
#define  PSZ_B4P  PSZ_B4_PORTRAIT
#define  PSZ_B4L  PSZ_B4_LANDSCAPE
#define  PSZ_A4P  PSZ_A4_PORTRAIT
#define  PSZ_A4L  PSZ_A4_LANDSCAPE
#define  PSZ_B5P  PSZ_B5_PORTRAIT
#define  PSZ_B5L  PSZ_B5_LANDSCAPE
#define  PSZ_A5P  PSZ_A5_PORTRAIT
#define  PSZ_A5L  PSZ_A5_LANDSCAPE
#define  PSZ_B6P  PSZ_B6_PORTRAIT
#define  PSZ_B6L  PSZ_B6_LANDSCAPE
#define  PSZ_A6P  PSZ_A6_PORTRAIT
#define  PSZ_A6L  PSZ_A6_LANDSCAPE

#define  PMG_REGULAR          0x00
#define  PMG_MIDWIDE          0x10
#define  PMG_WIDE             0x20
#define  PMG_EXTRAWIDE        0x30
#define  PMG_EXTRAEXTRAWIDE   0x40
#define  PMG_MIDNARROW        0x90
#define  PMG_NARROW           0xa0
#define  PMG_EXTRANARROW      0xb0
#define  PMG_EXTRAEXTRANARROW 0xc0

#define  PMG_REG  PMG_REGULAR
#define  PMG_MWD  PMG_MIDWIDE
#define  PMG_WID  PMG_WIDE
#define  PMG_XWD  PMG_EXTRAWIDE
#define  PMG_XXW  PMG_EXTRAEXTRAWIDE
#define  PMG_MNR  PMG_MIDNARROW
#define  PMG_NAR  PMG_NARROW
#define  PMG_XNR  PMG_EXTRANARROW
#define  PMG_XXN  PMG_EXTRAEXTRANARROW

typedef struct {
  PCSTR tit;
  int   flg;
  FREP  frep;
  UINT  pa;
} REPI;
typedef REPI * PREPI;
typedef const REPI * PCREPI;

HREP WINAPI CreateReport(PCREPI ri);
void WINAPI DeleteReport(HREP rp);

HWND WINAPI hWND(HREP rp);

void WINAPI SetReportInfo(HREP rp, PCSTR tit, PCSTR usr, int sr, int tm);
PCSTR WINAPI ReportTitle(HREP rp);
PCSTR WINAPI ReportAuthor(HREP rp);
int  WINAPI ReportDate(HREP rp);
int  WINAPI ReportTime(HREP rp);

void WINAPI SetReportExtra(HREP rp, UINT pa);
UINT WINAPI ReportExtra(HREP rp);

void WINAPI SetNPages(HREP rp, int npg);
int  WINAPI NPages(HREP rp);

void WINAPI SetPaperSize(HREP rp, REAL w, REAL h, REAL scl);
REAL WINAPI PaperWidth(HREP rp);
REAL WINAPI PaperHeight(HREP rp);

void WINAPI SetPageSize(HREP rp, REAL w, REAL h);
REAL WINAPI PageWidth(HREP rp);
REAL WINAPI PageHeight(HREP rp);
REAL WINAPI PageLeft(HREP rp);
REAL WINAPI PageTop(HREP rp);
REAL WINAPI PageRight(HREP rp);
REAL WINAPI PageBottom(HREP rp);

void WINAPI SetPageMargin(HREP rp, REAL lm, REAL tm, REAL rm, REAL bm);
REAL WINAPI BodyWidth(HREP rp);
REAL WINAPI BodyHeight(HREP rp);
REAL WINAPI BodyLeft(HREP rp);
REAL WINAPI BodyTop(HREP rp);
REAL WINAPI BodyRight(HREP rp);
REAL WINAPI BodyBottom(HREP rp);

void WINAPI SetHeaderHeight(HREP rp, REAL hh);
void WINAPI SetFooterHeight(HREP rp, REAL fh);
#define  HeaderWidth(rp)  BodyWidth(rp)
REAL WINAPI HeaderHeight(HREP rp);
#define  HeaderLeft(rp)   BodyLeft(rp)
REAL WINAPI HeaderTop(HREP rp);
#define  HeaderRight(rp)  BodyRight(rp)
#define  HeaderBottom(rp) BodyTop(rp)
#define  FooterWidth(rp)  BodyWidth(rp)
REAL WINAPI FooterHeight(HREP rp);
#define  FooterLeft(rp)   BodyLeft(rp)
#define  FooterTop(rp)    BodyBottom(rp)
#define  FooterRight(rp)  BodyRight(rp)
REAL WINAPI FooterBottom(HREP rp);

#define  RSW_AUTOSETTING    0x01
#define  RSW_SKIPPAGESETUP  0x02

#define  RSW_AUT  RSW_AUTOSETTING
#define  RSW_SKP  RSW_SKIPPAGESETUP

void WINAPI SetReqBodyWidth(HREP rp, REAL w, int sw);
REAL WINAPI ReqBodyWidth(HREP rp);

PCELL WINAPI PageCell(PCELL pc, HREP rp);
PCELL WINAPI BodyCell(PCELL pc, HREP rp);
PCELL WINAPI HeaderCell(PCELL pc, HREP rp);
PCELL WINAPI FooterCell(PCELL pc, HREP rp);

void WINAPI SetLineBreakTag(PCSTR lbt);
PCSTR WINAPI LineBreakTag(void);
void WINAPI SetPageBreakTag(PCSTR pbt);
PCSTR WINAPI PageBreakTag(void);
void WINAPI SetBrackets(PCSTR lbr, PCSTR rbr);
PCSTR WINAPI BracketL(void);
PCSTR WINAPI BracketR(void);
void WINAPI CloseBrackets(PCSTR *pp, BOOL sw);

void WINAPI SetSpacingFactor(HREP rp, REAL csf, REAL lsf);
REAL WINAPI SpacingFactorC(HREP rp);
REAL WINAPI SpacingFactorL(HREP rp);
PWHSZ WINAPI SpaceSize(PWHSZ sz, HREP rp);

typedef struct {
  char  face[32];
  REAL  ch;
  int   flg;
  BGRC  col;
  REAL  bph;
  REAL  aph;
  REAL  fiw;
  REAL  liw;
  REAL  riw;
} PFMI;
typedef PFMI * PPFMI;
typedef const PFMI * PCPFMI;

BOOL WINAPI RegisterParagraphFormat(int id, PCSTR tag, PCPFMI fi);

REAL WINAPI ParagraphSpacingB(HREP rp);
REAL WINAPI ParagraphSpacingA(HREP rp);
int  WINAPI ParagraphAlignment(HREP rp);
REAL WINAPI ParagraphIndentationF(HREP rp);
REAL WINAPI ParagraphIndentationL(HREP rp);
REAL WINAPI ParagraphIndentationR(HREP rp);

#define  STA_ENDOFLINE       0x10
#define  STA_ENDOFPARAGRAPH  0x20
#define  STA_ENDOFPAGE       0x40

#define  STA_LIN  STA_ENDOFLINE
#define  STA_PAR  STA_ENDOFPARAGRAPH
#define  STA_END  STA_ENDOFPAGE

void WINAPI SetFlowTextStatus(HREP rp, int sta);
int  WINAPI FlowTextStatus(HREP rp);
void WINAPI StartFlowText(HREP rp);

typedef void (CALLBACK * FPAR)(HREP rp, PCSTR *pp);
typedef void (CALLBACK * FCHR)(HREP rp, PCSTR *pp, PCCELL pc, PWHSZ sz);

typedef struct {
  PCSTR nam;
  FPAR  fpar;
  FCHR  fchr;
} FTEI;
typedef FTEI * PFTEI;
typedef const FTEI * PCFTEI;

typedef struct {
  PFTEI pi;
  int   ni;
} FTES;
typedef FTES * PFTES;
typedef const FTES * PCFTES;

BOOL WINAPI RegisterFlowTextTags(PCFTES ps);

void WINAPI FlowText(HREP rp, PCSTR *pp, PCCELL pc);

PPSDS WINAPI PageSetupInfo(HREP rp);

void WINAPI StartPrint(HREP rp, HDC hdc, int dpi);
void WINAPI EndPrint(HREP rp);
void WINAPI DrawPage(HREP rp, int pg);
void WINAPI SkipPage(HREP rp, int pg);

BOOL WINAPI ReportSetup(HWND hwnd, HREP rp);
BOOL WINAPI PageSetup(HWND hwnd, HREP rp);
BOOL WINAPI PrintReport(HWND hwnd, HREP rp);

// ---------------------------------------------------------------------- //

UINT WINAPI BOLFROVer(void);

BOOL WINAPI InitBOLFRO(UINT pa);
void WINAPI TermBOLFRO(void);

#ifdef __cplusplus
}
#endif

// ---------------------------------------------------------------------- //
